#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['VectorDimentionConflictError', 'VectorBase', 'CoordinateVectorDimentionsAreNotMatch', 'CoordinateVectorNd', 'VectorNd', 'DirectedGraphNd']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from time import perf_counter
from cengal.entities.copyable import *
from cengal.math.geometry.point import *
from typing import Any, FrozenSet, Optional, Sequence, List, Set, Tuple, Dict, Union
from math import sqrt
numpy_present = True
try:
    import numpy as np
    from numpy import linalg, array
except:
    numpy_present = False


class VectorDimentionConflictError(ValueError):
    pass


class VectorBase(CopyableMixin):
    def copy(self) -> 'VectorBase':
        raise NotImplementedError

    def shallow_copy(self) -> 'VectorBase':
        raise NotImplementedError
    
    def updated_copy(self, update: Dict) -> 'VectorBase':
        raise NotImplementedError

    @property
    def dim(self) -> int:
        raise NotImplementedError
    
    def _is_dimension_ok(self) -> bool:
        raise NotImplementedError

    def _is_new_points_dimension_ok(self, new_point: PointNd) -> bool:
        raise NotImplementedError

    def _set_forgotten_points_to_zero(self):
        raise NotImplementedError

    def __call__(self
            , *args: Tuple[PointNd]
            , **kwargs: Dict[str, PointNd]
            ) -> PointNd:
        raise NotImplementedError
    
    def __len__(self):
        raise NotImplementedError


class CoordinateVectorDimentionsAreNotMatch(ValueError):
    pass


class CoordinateVectorNd(VectorBase):
    _terminal_point_names = {'tp', 'terminal_point', 'terminal point'}

    def __init__(self, terminal_point: PointNd) -> None:
        self.terminal_point: PointNd = terminal_point
    
    def __repr__(self):
        return f'{type(self).__name__}(tp={repr(self.terminal_point)})'
    
    def copy(self) -> 'CoordinateVectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['terminal_point'] = self.terminal_point.copy()
        return result
    
    def shallow_copy(self) -> 'CoordinateVectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['terminal_point'] = self.terminal_point
        return result
    
    def updated_copy(self, update: Dict) -> 'CoordinateVectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['terminal_point'] = get_dict_key_with_callable_default(update, 'terminal_point', lambda: self.terminal_point.copy())
        return result
    
    @property
    def dim(self):
        return self.terminal_point.dim
    
    def _is_dimension_ok(self) -> bool:
        return True

    def __call__(self
            , *args: Tuple[PointNd]
            , **kwargs: Dict[str, PointNd]
            ) -> PointNd:
        if args:
            first_item = args[0]
            if isinstance(first_item, CoordinateVectorNd):
                data = first_item()
            elif isinstance(first_item, PointNd):
                data = first_item
            else:
                raise ValueError
            
            self.terminal_point = data
        elif kwargs:
            terminal_point = kwargs.get('terminal_point')
            for terminal_point_name_variant in self._terminal_point_names:
                terminal_point = kwargs.get(terminal_point_name_variant, None)
                if terminal_point is not None:
                    self.terminal_point = terminal_point
        
        return self.terminal_point
    
    def __getitem__(self, index):
        if isinstance(index, int):
            if 0 == index:
                return self.terminal_point
            else:
                raise IndexError
        else:
            if index in self._terminal_point_names:
                return self.terminal_point
            else:
                raise KeyError
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            if 0 == index:
                self.terminal_point = value
            else:
                raise IndexError
        else:
            if index in self._terminal_point_names:
                self.terminal_point = value
            else:
                raise KeyError
    
    def __len__(self):
        return 1

    def magnitude(self) -> float:
        if numpy_present:
            return linalg.norm(self.terminal_point())
        else:
            result = 0
            for item in self.terminal_point:
                result += item * item
            
            return sqrt(result)
    
    def length(self) -> float:
        return self.magnitude()

    def magnitude_square(self) -> float:
        if numpy_present:
            tp = self.terminal_point()
            return sum(tp * tp)
        else:
            result = 0
            for item in self.terminal_point:
                result += item * item
            
            return result
    
    # add
    def __add__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] + other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] + other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __radd__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        return self.__add__(other)
    
    def __iadd__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if numpy_present:
            if isinstance(other, CoordinateVectorNd):
                self.terminal_point._point += other.terminal_point()
            elif isinstance(other, (float, int)):
                self.terminal_point._point += other
            else:
                raise NotImplemented
        else:
            if isinstance(other, CoordinateVectorNd):
                for index in range(self.dim):
                    self.terminal_point[index] += other.terminal_point[index]
            elif isinstance(other, (float, int)):
                for index in range(self.dim):
                    self.terminal_point[index] += other
            else:
                raise NotImplemented

        return self
    
    # sub
    def __sub__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if numpy_present:
            result = None
            if isinstance(other, CoordinateVectorNd):
                result = self.terminal_point() - other.terminal_point()
            elif isinstance(other, (float, int)):
                result = self.terminal_point() - other
            else:
                raise NotImplemented
            
            return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result, ndarray_type=type(result[0].item()))))
        else:
            result = list()
            if isinstance(other, CoordinateVectorNd):
                for index in range(self.dim):
                    result.append(self.terminal_point[index] - other.terminal_point[index])
            elif isinstance(other, (float, int)):
                for index in range(self.dim):
                    result.append(self.terminal_point[index] - other)
            else:
                raise NotImplemented
            
            return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rsub__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] - self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other - self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __isub__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] -= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] -= other
        else:
            raise NotImplemented

        return self
    
    # mul
    def __mul__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        """Multiplication to number or to other coordinate vector (item by item)

        v1 = [1, 2, 3]
        v2 = [2, 3, 1]
        v3 = v1 * v2 = [1 * 2, 2 * 3, 3 * 1] = [2, 6, 3]

        Args:
            other (Union[float, int, &#39;CoordinateVectorNd&#39;]): _description_

        Raises:
            NotImplemented: _description_

        Returns:
            CoordinateVectorNd: _description_
        """
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] * other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] * other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rmul__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        """Multiplication to number or to other coordinate vector (item by item)

        v1 = [1, 2, 3]
        v2 = [2, 3, 1]
        v3 = v1 * v2 = [1 * 2, 2 * 3, 3 * 1] = [2, 6, 3]

        Args:
            other (Union[float, int, &#39;CoordinateVectorNd&#39;]): _description_

        Returns:
            CoordinateVectorNd: _description_
        """
        return self.__mul__(other)
    
    def __imul__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        """Multiplication to number or to other coordinate vector (item by item)

        v1 = [1, 2, 3]
        v2 = [2, 3, 1]
        v3 = v1 * v2 = [1 * 2, 2 * 3, 3 * 1] = [2, 6, 3]

        Args:
            other (Union[float, int, &#39;CoordinateVectorNd&#39;]): _description_

        Raises:
            NotImplemented: _description_

        Returns:
            CoordinateVectorNd: _description_
        """
        if numpy_present:
            if isinstance(other, CoordinateVectorNd):
                self.terminal_point._point *= other.terminal_point()
            elif isinstance(other, (float, int)):
                self.terminal_point._point *= other
            else:
                raise NotImplemented
        else:
            if isinstance(other, CoordinateVectorNd):
                for index in range(self.dim):
                    self.terminal_point[index] *= other.terminal_point[index]
            elif isinstance(other, (float, int)):
                for index in range(self.dim):
                    self.terminal_point[index] *= other
            else:
                raise NotImplemented

        return self
    
    # matmul
    def __matmul__(self, other: Union[float, int, 'CoordinateVectorNd']) -> Union[float, int]:
        """Dot product
        >> cv0 = CoordinateVectorNd(2, 4)
        >> cv1 = CoordinateVectorNd(1, 3)
        >> dot_product = cv0 @ cv1
        >> print(dot_product)
        14

        Args:
            other (Union[float, int, &#39;CoordinateVectorNd&#39;]): _description_

        Raises:
            NotImplemented: _description_
            CoordinateVectorDimentionsAreNotMatch: _description_

        Returns:
            Union[float, int]: _description_
        """
        if not isinstance(other, CoordinateVectorNd):
            raise NotImplemented
        
        dim = self.dim
        if dim != other.dim:
            raise CoordinateVectorDimentionsAreNotMatch

        result = 0
        for index in range(dim):
            result += self.terminal_point[index] * other.terminal_point[index]
        
        return result
    
    # truediv
    def __truediv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] / other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] / other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rtruediv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] / self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other / self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __itruediv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if numpy_present:
            if isinstance(other, CoordinateVectorNd):
                self.terminal_point._point /= other.terminal_point()
            elif isinstance(other, (float, int)):
                self.terminal_point._point /= other
            else:
                raise NotImplemented
        else:
            if isinstance(other, CoordinateVectorNd):
                for index in range(self.dim):
                    self.terminal_point[index] /= other.terminal_point[index]
            elif isinstance(other, (float, int)):
                for index in range(self.dim):
                    self.terminal_point[index] /= other
            else:
                raise NotImplemented

        return self
    
    # floordiv
    def __floordiv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] // other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] // other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rfloordiv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] // self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other // self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ifloordiv__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if numpy_present:
            if isinstance(other, CoordinateVectorNd):
                self.terminal_point._point //= other.terminal_point()
            elif isinstance(other, (float, int)):
                self.terminal_point._point //= other
            else:
                raise NotImplemented
        else:
            if isinstance(other, CoordinateVectorNd):
                for index in range(self.dim):
                    self.terminal_point[index] //= other.terminal_point[index]
            elif isinstance(other, (float, int)):
                for index in range(self.dim):
                    self.terminal_point[index] //= other
            else:
                raise NotImplemented

        return self
    
    # mod
    def __mod__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] % other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] % other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rmod__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] % self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other % self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __imod__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] %= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] %= other
        else:
            raise NotImplemented

        return self
    
    # pow
    def __pow__(self, other: Union[float, int, 'CoordinateVectorNd'], modulo) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(pow(self.terminal_point[index], other.terminal_point[index], modulo))
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(pow(self.terminal_point[index], other, modulo))
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rpow__(self, other: Union[float, int, 'CoordinateVectorNd'], modulo) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(pow(other.terminal_point[index], self.terminal_point[index], modulo))
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(pow(other, self.terminal_point[index], modulo))
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ipow__(self, other: Union[float, int, 'CoordinateVectorNd'], modulo) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index].__ipow__(other.terminal_point[index], modulo)
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index].__ipow__(other, modulo)
        else:
            raise NotImplemented

        return self
    
    # lshift
    def __lshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] << other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] << other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rlshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] << self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other << self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ilshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] <<= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] <<= other
        else:
            raise NotImplemented

        return self
    
    # rshift
    def __rshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] >> other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] >> other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rrshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] >> self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other >> self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __irshift__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] >>= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] >>= other
        else:
            raise NotImplemented

        return self
    
    # and
    def __and__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] & other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] & other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rand__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] & self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other & self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __iand__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] &= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] &= other
        else:
            raise NotImplemented

        return self
    
    # xor
    def __xor__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] ^ other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] ^ other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __rxor__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] ^ self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other ^ self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ixor__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] ^= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] ^= other
        else:
            raise NotImplemented

        return self
    
    # or
    def __or__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(self.terminal_point[index] | other.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(self.terminal_point[index] | other)
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ror__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        result = list()
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                result.append(other.terminal_point[index] | self.terminal_point[index])
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                result.append(other | self.terminal_point[index])
        else:
            raise NotImplemented
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __ior__(self, other: Union[float, int, 'CoordinateVectorNd']) -> 'CoordinateVectorNd':
        if isinstance(other, CoordinateVectorNd):
            for index in range(self.dim):
                self.terminal_point[index] |= other.terminal_point[index]
        elif isinstance(other, (float, int)):
            for index in range(self.dim):
                self.terminal_point[index] |= other
        else:
            raise NotImplemented

        return self
    
    # neg
    def __neg__(self) -> 'CoordinateVectorNd':
        if numpy_present:
            result = self.terminal_point().__neg__()
            return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result, ndarray_type=type(result[0].item()))))
        else:
            result = list()
            for index in range(self.dim):
                result.append(self.terminal_point[index].__neg__())
            
            return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # pos
    def __pos__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__pos__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # abs
    def __abs__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__abs__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # invert
    def __invert__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__invert__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # complex
    def __complex__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__complex__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # int
    def __int__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__int__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # float
    def __float__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__float__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # round
    def __round__(self, ndigits) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__round__(ndigits))
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # trunc
    def __trunc__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__trunc__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # floor
    def __floor__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__floor__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    # ceil
    def __ceil__(self) -> 'CoordinateVectorNd':
        result = list()
        for index in range(self.dim):
            result.append(self.terminal_point[index].__ceil__())
        
        return CoordinateVectorNd(convert_point_to_xyz(PointNd(len(result), result)))
    
    def __eq__(self, other) -> bool:
        if isinstance(other, PointNd):
            return self.terminal_point == other.terminal_point
        
        return False
    
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)
    
    def __hash__(self) -> int:
        return hash(self.terminal_point)
    
    def __bool__(self) -> bool:
        """If not zero point

        Returns:
            bool: _description_
        """
        return self.terminal_point.__bool__()

    # TODO: Другие операции умножения применимые к координатным векторам - реализовать в методах.
    # А неприменимые - в соответствующих классах (например DirectedGraphNd отлично подходит на роль 
    # матрицы с которой будут производится манипуляции. Каждая точка - отдельный столбец. 
    # Направление слева-направо: слева нулевой индекс, а справа - максимальный индекс)


class VectorNd(VectorBase):
    _initial_point_names = {'ip', 'initial_point', 'initial point'}
    _terminal_point_names = {'tp', 'terminal_point', 'terminal point'}

    def __init__(self
            , *args: Tuple[PointNd]
            , **kwargs: Dict[str, PointNd]
            ) -> None:
        self.initial_point: PointNd = None
        self.terminal_point: PointNd = None
        self(*args, **kwargs)
    
    def copy(self) -> 'VectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['initial_point'] = self.initial_point.copy()
        result.__dict__['terminal_point'] = self.terminal_point.copy()
        return result
    
    def shallow_copy(self) -> 'VectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['initial_point'] = self.initial_point
        result.__dict__['terminal_point'] = self.terminal_point
        return result
    
    def updated_copy(self, update: Dict) -> 'VectorNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['initial_point'] = get_dict_key_with_callable_default(update, 'initial_point', lambda: self.initial_point.copy())
        result.__dict__['terminal_point'] = get_dict_key_with_callable_default(update, 'terminal_point', lambda: self.terminal_point.copy())
        return result
    
    @property
    def dim(self):
        if self.initial_point is None:
            return None
        
        return self.initial_point.dim

    def _is_dimension_ok(self) -> bool:
        if (self.initial_point is None) and (self.terminal_point is None):
            return True
        
        initial_dim = self.initial_point.dim
        terminal_dim = self.terminal_point.dim
        if initial_dim == terminal_dim:
            return True
        else:
            raise VectorDimentionConflictError
    
    def _set_forgotten_points_to_zero(self):
        if (self.initial_point is not None) and (self.terminal_point is None):
            dim = self.initial_point.dim
            self.terminal_point = PointNd(dim, [0] * dim)
        elif (self.initial_point is None) and (self.terminal_point is not None):
            dim = self.terminal_point.dim
            self.initial_point = PointNd(dim, [0] * dim)
        elif (self.initial_point is None) and (self.terminal_point is None):
            pass

    def __call__(self
            , *args: Tuple[PointNd]
            , **kwargs: Dict[str, PointNd]
            ) -> Tuple[PointNd]:
        if args:
            first_item = args[0]
            data = first_item
            if isinstance(first_item, VectorNd):
                data = first_item()
            if isinstance(first_item, CoordinateVectorNd):
                terminal_point = first_item()
                data = [PointNd(terminal_point.dim), terminal_point]
            elif isinstance(first_item, PointNd):
                data = args
            
            self.initial_point = data[0]
            data_len = len(data)
            if data_len == 2:
                self.terminal_point = data[1]
            elif data_len > 2:
                raise IndexError
            
            self._set_forgotten_points_to_zero()
            self._is_dimension_ok()
        elif kwargs:
            for initial_point_name_variant in self._initial_point_names:
                initial_point = kwargs.get(initial_point_name_variant, None)
                if initial_point is not None:
                    self.initial_point = initial_point
                    break

            for terminal_point_name_variant in self._terminal_point_names:
                terminal_point = kwargs.get(terminal_point_name_variant, None)
                if terminal_point is not None:
                    self.terminal_point = terminal_point
                    break
            
            self._set_forgotten_points_to_zero()
            self._is_dimension_ok()
        
        return [self.initial_point, self.terminal_point]
    
    def __getitem__(self, index):
        if isinstance(index, int):
            if 0 == index:
                return self.initial_point
            elif 1 == index:
                return self.terminal_point
            else:
                raise IndexError
        else:
            if index in self._initial_point_names:
                return self.initial_point
            elif index in self._terminal_point_names:
                return self.terminal_point
            else:
                raise KeyError
    
    def __setitem__(self, index, value):
        if isinstance(index, int):
            if 0 == index:
                self.initial_point = value
            elif 1 == index:
                self.terminal_point = value
            else:
                raise IndexError
        else:
            if index in self._initial_point_names:
                self.initial_point = value
            elif index in self._terminal_point_names:
                self.terminal_point = value
            else:
                raise KeyError
        
        self._set_forgotten_points_to_zero()
        self._is_dimension_ok()
    
    def __len__(self):
        return 2

    def magnitude(self) -> float:
        if numpy_present:
            distance_vec = self.terminal_point() - self.initial_point()
            return linalg.norm(distance_vec)
        else:
            init_vec: CoordinateVectorNd = CoordinateVectorNd(self.initial_point)
            term_vec: CoordinateVectorNd = CoordinateVectorNd(self.terminal_point)
            distance_vec: CoordinateVectorNd = term_vec - init_vec
            return distance_vec.magnitude()
    
    def length(self) -> float:
        return self.magnitude()

    def magnitude_square(self) -> float:
        if numpy_present:
            distance_vec = self.terminal_point() - self.initial_point()
            return sum(distance_vec * distance_vec)
        else:
            init_vec: CoordinateVectorNd = CoordinateVectorNd(self.initial_point)
            term_vec: CoordinateVectorNd = CoordinateVectorNd(self.terminal_point)
            distance_vec: CoordinateVectorNd = term_vec - init_vec
            return distance_vec.magnitude_square()
    
    def is_convergence(self, other: 'VectorNd'):
        if numpy_present:
            init_dist = other.initial_point() - self.initial_point()
            term_dist = other.terminal_point() - self.terminal_point()
            return sum(init_dist * init_dist) > sum(term_dist * term_dist)
        else:
            return VectorNd(self.initial_point, other.initial_point).magnitude_square() \
                > VectorNd(self.terminal_point, other.terminal_point).magnitude_square()
    
    def is_divergence(self, other: 'VectorNd'):
        if numpy_present:
            init_dist = other.initial_point() - self.initial_point()
            term_dist = other.terminal_point() - self.terminal_point()
            return sum(init_dist * init_dist) < sum(term_dist * term_dist)
        else:
            return VectorNd(self.initial_point, other.initial_point).magnitude_square() \
                < VectorNd(self.terminal_point, other.terminal_point).magnitude_square()
    
    def is_conver_diver_parity(self, other: 'VectorNd'):
        if numpy_present:
            init_dist = other.initial_point() - self.initial_point()
            term_dist = other.terminal_point() - self.terminal_point()
            return sum(init_dist * init_dist) == sum(term_dist * term_dist)
        else:
            return VectorNd(self.initial_point, other.initial_point).magnitude_square() \
                == VectorNd(self.terminal_point, other.terminal_point).magnitude_square()
    
    def conver_diver_state(self, other: 'VectorNd') -> int:
        if numpy_present:
            init_dist = other.initial_point() - self.initial_point()
            term_dist = other.terminal_point() - self.terminal_point()
            init_magnitude_square = sum(init_dist * init_dist)
            term_magnitude_square = sum(term_dist * term_dist)
        else:
            init_magnitude_square = VectorNd(self.initial_point, other.initial_point).magnitude_square()
            term_magnitude_square = VectorNd(self.terminal_point, other.terminal_point).magnitude_square()

        if init_magnitude_square > term_magnitude_square:
            return 1
        elif init_magnitude_square == term_magnitude_square:
            return 0
        elif init_magnitude_square < term_magnitude_square:
            return -1


class DirectedGraphNd(VectorBase):
    def __init__(self
            , *args: Union[Tuple[PointNd], Tuple[Sequence[PointNd]]]
            ) -> None:
        self.points_list: List[PointNd] = None
        self(*args)
    
    def copy(self) -> 'DirectedGraphNd':
        cls = self.__class__
        result = cls.__new__(cls)
        if self.points_list is None:
            result.__dict__['points_list'] = self.points_list
        else:
            result.__dict__['points_list'] = [point.copy() for point in self.points_list]

        return result
    
    def shallow_copy(self) -> 'DirectedGraphNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['points_list'] = self.points_list
        return result
    
    def updated_copy(self, update: Dict) -> 'DirectedGraphNd':
        cls = self.__class__
        result = cls.__new__(cls)
        def copy_points():
            if self.points_list is None:
                return self.points_list
            else:
                return [point.copy() for point in self.points_list]

        result.__dict__['points_list'] = get_dict_key_with_callable_default(update, 'points_list', copy_points)
        return result
    
    @property
    def dim(self):
        if self.points_list:
            return self.points_list[0].dim
        else:
            return None

    def _is_dimension_ok(self) -> bool:
        if self.points_list:
            initial_dim = self.points_list[0].dim
            for point in self.points_list:
                if initial_dim != point.dim:
                    raise VectorDimentionConflictError
        else:
            return True

    def _is_new_points_dimension_ok(self, new_point: PointNd) -> bool:
        dim = self.dim
        if dim is None:
            return True
        else:
            if dim == new_point.dim:
                return True
            else:
                raise VectorDimentionConflictError

    def __call__(self
            , *args: Union[Tuple[PointNd], Tuple[Sequence[PointNd]]]
            ) -> Optional[List[PointNd]]:
        if args:
            first_item = args[0]
            data = first_item
            if isinstance(first_item, VectorNd):
                data = first_item()
            elif isinstance(first_item, PointNd):
                data = args
            
            self.points_list = list(data)
        
        self._is_dimension_ok()
        return self.points_list
    
    def get_vector(self, index: int) -> VectorNd:
        if 0 <= index < (len(self) - 1):
            return VectorNd(self[index], self[index + 1])
        else:
            raise IndexError
    
    def get_vector_reversed(self, index: int) -> VectorNd:
        if 1 <= index < len(self):
            return VectorNd(self[index], self[index - 1])
        else:
            raise IndexError
    
    def __getitem__(self, index):
        return self.points_list[index]
    
    def __setitem__(self, index, value):
        self._is_new_points_dimension_ok(value)
        self.points_list[index] = value
    
    def __len__(self):
        return len(self.points_list)


# class VectorBase:
#     def move(self, point: PointBase):
#         raise NotImplementedError

# class Vector1d:
#     pass

# class Vector3d:
#     pass



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['S_OK', 'WH_CALLWNDPROC', 'WH_CALLWNDPROCRET', 'WCA_ACCENT_POLICY', 'WS_BORDER', 'WS_DLGFRAME', 'WS_OVERLAPPED', 'WS_THICKFRAME', 'WS_CAPTION', 'WS_SYSMENU', 'WS_MINIMIZEBOX', 'WS_MAXIMIZEBOX', 'WS_POPUP', 'WS_OVERLAPPEDWINDOW', 'SM_CXPADDEDBORDER', 'KnownfolderidConstants']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

from ctypes import c_int
from cengal.ctypes_tools.types import GUID
from uuid import UUID

S_OK = 0

WH_CALLWNDPROC = c_int(4)
WH_CALLWNDPROCRET = c_int(12)


WCA_ACCENT_POLICY = 19

WS_BORDER = 0x00800000
WS_DLGFRAME = 0x00400000
WS_OVERLAPPED = 0x00000000
WS_THICKFRAME = 0x00040000
WS_CAPTION = 0x00C00000  # WS_BORDER | WS_DLGFRAME
WS_SYSMENU = 0x00080000
WS_MINIMIZEBOX = 0x00020000
WS_MAXIMIZEBOX = 0x00010000
WS_POPUP = 0x80000000
WS_OVERLAPPEDWINDOW = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX


SM_CXPADDEDBORDER = 92

# SW_MAXIMIZE: int = 3


# MONITOR_DEFAULTTONULL = 0x00000000


# WM_NCCREATE = 0x0081
# WM_NCCALCSIZE = 0x0083
# WM_NCHITTEST = 0x0084
# WM_NCACTIVATE = 0x0086
# WM_CLOSE = 0x0010
# WM_DESTROY = 0x0002
# WM_KEYDOWN = 0x0100
# WM_SYSKEYDOWN = 0x0104


# GWL_WNDPROC = -4

# CS_HREDRAW = 0x0002
# CS_VREDRAW = 0x0001


# CW_USEDEFAULT = 0x80000000


# VK_F8 = 0x77
# VK_F9 = 0x78
# VK_F10 = 0x79
# VK_F11 = 0x7A


class KnownfolderidConstants:
    FOLDERID_AccountPictures = GUID.from_uuid(UUID('{008ca0b1-55b4-4c56-b8a8-4de4b299d3be}'))
    FOLDERID_AddNewPrograms = GUID.from_uuid(UUID('{de61d971-5ebc-4f02-a3a9-6c82895e5c04}'))
    FOLDERID_AdminTools = GUID.from_uuid(UUID('{724EF170-A42D-4FEF-9F26-B60E846FBA4F}'))
    FOLDERID_ApplicationShortcuts = GUID.from_uuid(UUID('{A3918781-E5F2-4890-B3D9-A7E54332328C}'))
    FOLDERID_CDBurning = GUID.from_uuid(UUID('{9E52AB10-F80D-49DF-ACB8-4330F5687855}'))
    FOLDERID_CommonAdminTools = GUID.from_uuid(UUID('{D0384E7D-BAC3-4797-8F14-CBA229B392B5}'))
    FOLDERID_CommonAppData = GUID.from_uuid(UUID('{62AB5D82-FDC1-4DC3-A9DD-070D1D495D97}'))
    FOLDERID_CommonDesktopDirectory = GUID.from_uuid(UUID('{C4AA340D-F20F-4863-AFEF-F87EF2E6BA25}'))
    FOLDERID_CommonDocuments = GUID.from_uuid(UUID('{ED4824AF-DCE4-45A8-81E2-FC7965083634}'))
    FOLDERID_CommonDownloads = GUID.from_uuid(UUID('{f7ce2e13-8c90-4e1f-ba7e-1925aebd977c}'))
    FOLDERID_CommonMusic = GUID.from_uuid(UUID('{4BD8D571-6D19-48D3-BE97-422220080E43}'))
    FOLDERID_CommonOemLinks = GUID.from_uuid(UUID('{C1BAE2D0-10DF-4334-BEDD-7AA20B227A9D}'))
    FOLDERID_CommonPictures = GUID.from_uuid(UUID('{B6EBFB86-6907-413C-9AF7-4FC2ABF07CC5}'))
    FOLDERID_CommonPrograms = GUID.from_uuid(UUID('{0139D44E-6AFE-49F2-8690-3DAFCAE6FFB8}'))
    FOLDERID_CommonStartMenu = GUID.from_uuid(UUID('{A4115719-D62E-491D-AA7C-E74B8BE3B067}'))
    FOLDERID_CommonStartup = GUID.from_uuid(UUID('{82A5EA35-D9CD-47C5-9629-E15D2F714E6E}'))
    FOLDERID_CommonTemplates = GUID.from_uuid(UUID('{B94237E7-57AC-4347-9151-B08C6C32D1F7}'))
    FOLDERID_CommonVideos = GUID.from_uuid(UUID('{18989B1D-99B5-455B-841C-AB7C74E4DDFC}'))
    FOLDERID_ComputerFolder = GUID.from_uuid(UUID('{0AC0837C-BBF8-452A-850D-79D08E667CA7}'))
    FOLDERID_ConflictFolder = GUID.from_uuid(UUID('{4bfefb45-347d-4006-a5be-ac0cb0567192}'))
    FOLDERID_ConnectionsFolder = GUID.from_uuid(UUID('{6F0CD92B-2E97-45D1-88FF-B0D186B8DEDD}'))
    FOLDERID_Contacts = GUID.from_uuid(UUID('{56784854-C6CB-462b-8169-88E350ACB882}'))
    FOLDERID_ControlPanelFolder = GUID.from_uuid(UUID('{82A74AEB-AEB4-465C-A014-D097EE346D63}'))
    FOLDERID_Cookies = GUID.from_uuid(UUID('{2B0F765D-C0E9-4171-908E-08A611B84FF6}'))
    FOLDERID_Desktop = GUID.from_uuid(UUID('{B4BFCC3A-DB2C-424C-B029-7FE99A87C641}'))
    FOLDERID_DeviceMetadataStore = GUID.from_uuid(UUID('{5CE4A5E9-E4EB-479D-B89F-130C02886155}'))
    FOLDERID_Documents = GUID.from_uuid(UUID('{FDD39AD0-238F-46AF-ADB4-6C85480369C7}'))
    FOLDERID_DocumentsLibrary = GUID.from_uuid(UUID('{7B0DB17D-9CD2-4A93-9733-46CC89022E7C}'))
    FOLDERID_Downloads = GUID.from_uuid(UUID('{374DE290-123F-4565-9164-39C4925E467B}'))
    FOLDERID_Favorites = GUID.from_uuid(UUID('{1777F761-68AD-4D8A-87BD-30B759FA33DD}'))
    FOLDERID_Fonts = GUID.from_uuid(UUID('{FD228CB7-AE11-4AE3-864C-16F3910AB8FE}'))
    FOLDERID_GameTasks = GUID.from_uuid(UUID('{054FAE61-4DD8-4787-80B6-090220C4B700}'))
    FOLDERID_History = GUID.from_uuid(UUID('{D9DC8A3B-B784-432E-A781-5A1130A75963}'))
    FOLDERID_HomeGroup = GUID.from_uuid(UUID('{52528A6B-B9E3-4ADD-B60D-588C2DBA842D}'))
    FOLDERID_HomeGroupCurrentUser = GUID.from_uuid(UUID('{9B74B6A3-0DFD-4f11-9E78-5F7800F2E772}'))
    FOLDERID_ImplicitAppShortcuts = GUID.from_uuid(UUID('{BCB5256F-79F6-4CEE-B725-DC34E402FD46}'))
    FOLDERID_InternetCache = GUID.from_uuid(UUID('{352481E8-33BE-4251-BA85-6007CAEDCF9D}'))
    FOLDERID_InternetFolder = GUID.from_uuid(UUID('{4D9F7874-4E0C-4904-967B-40B0D20C3E4B}'))
    FOLDERID_Libraries = GUID.from_uuid(UUID('{1B3EA5DC-B587-4786-B4EF-BD1DC332AEAE}'))
    FOLDERID_Links = GUID.from_uuid(UUID('{bfb9d5e0-c6a9-404c-b2b2-ae6db6af4968}'))
    FOLDERID_LocalAppData = GUID.from_uuid(UUID('{F1B32785-6FBA-4FCF-9D55-7B8E7F157091}'))
    FOLDERID_LocalAppDataLow = GUID.from_uuid(UUID('{A520A1A4-1780-4FF6-BD18-167343C5AF16}'))
    FOLDERID_LocalizedResourcesDir = GUID.from_uuid(UUID('{2A00375E-224C-49DE-B8D1-440DF7EF3DDC}'))
    FOLDERID_Music = GUID.from_uuid(UUID('{4BD8D571-6D19-48D3-BE97-422220080E43}'))
    FOLDERID_MusicLibrary = GUID.from_uuid(UUID('{2112AB0A-C86A-4FFE-A368-0DE96E47012E}'))
    FOLDERID_NetHood = GUID.from_uuid(UUID('{C5ABBF53-E17F-4121-8900-86626FC2C973}'))
    FOLDERID_NetworkFolder = GUID.from_uuid(UUID('{D20BEEC4-5CA8-4905-AE3B-BF251EA09B53}'))
    FOLDERID_OriginalImages = GUID.from_uuid(UUID('{2C36C0AA-5812-4b87-BFD0-4CD0DFB19B39}'))
    FOLDERID_PhotoAlbums = GUID.from_uuid(UUID('{69D2CF90-FC33-4FB7-9A0C-EBB0F0FCB43C}'))
    FOLDERID_Pictures = GUID.from_uuid(UUID('{33E28130-4E1E-4676-835A-98395C3BC3BB}'))
    FOLDERID_PicturesLibrary = GUID.from_uuid(UUID('{A990AE9F-A03B-4E80-94BC-9912D7504104}'))
    FOLDERID_Playlists = GUID.from_uuid(UUID('{DE92C1C7-837F-4F69-A3BB-86E631204A23}'))
    FOLDERID_PrintHood = GUID.from_uuid(UUID('{9274BD8D-CFD1-41C3-B35E-B13F55A758F4}'))
    FOLDERID_PrintersFolder = GUID.from_uuid(UUID('{76FC4E2D-D6AD-4519-A663-37BD56068185}'))
    FOLDERID_Profile = GUID.from_uuid(UUID('{5E6C858F-0E22-4760-9AFE-EA3317B67173}'))
    FOLDERID_ProgramData = GUID.from_uuid(UUID('{62AB5D82-FDC1-4DC3-A9DD-070D1D495D97}'))
    FOLDERID_ProgramFiles = GUID.from_uuid(UUID('{905e63b6-c1bf-494e-b29c-65b732d3d21a}'))
    FOLDERID_ProgramFilesX64 = GUID.from_uuid(UUID('{6D809377-6AF0-444b-8957-A3773F02200E}'))
    FOLDERID_ProgramFilesX86 = GUID.from_uuid(UUID('{7C5A40EF-A0FB-4BFC-874A-C0F2E0B9FA8E}'))
    FOLDERID_ProgramFilesCommon = GUID.from_uuid(UUID('{F7F1ED05-9F6D-47A2-AAAE-29D317C6F066}'))
    FOLDERID_ProgramFilesCommonX64 = GUID.from_uuid(UUID('{6365D5A7-0F0D-45E5-87F6-0DA56B6A4F7D}'))
    FOLDERID_ProgramFilesCommonX86 = GUID.from_uuid(UUID('{DE974D24-D9C6-4D3E-BF91-F4455120B917}'))
    FOLDERID_Programs = GUID.from_uuid(UUID('{A77F5D77-2E2B-44C3-A6A2-ABA601054A51}'))
    FOLDERID_Public = GUID.from_uuid(UUID('{DFDF76A2-C82A-4D63-906A-5644AC457385}'))
    FOLDERID_PublicDesktop = GUID.from_uuid(UUID('{C4AA340D-F20F-4863-AFEF-F87EF2E6BA25}'))
    FOLDERID_PublicDocuments = GUID.from_uuid(UUID('{ED4824AF-DCE4-45A8-81E2-FC7965083634}'))
    FOLDERID_PublicDownloads = GUID.from_uuid(UUID('{3D644C9B-1FB8-4f30-9B45-F670235F79C0}'))
    FOLDERID_PublicGameTasks = GUID.from_uuid(UUID('{DEBF2536-E1A8-4c59-B6A2-414586476AEA}'))
    FOLDERID_PublicLibraries = GUID.from_uuid(UUID('{48DAF80B-E6CF-4F4E-B800-0E69D84EE384}'))
    FOLDERID_PublicMusic = GUID.from_uuid(UUID('{3214FAB5-9757-4298-BB61-92A9DEAA44FF}'))
    FOLDERID_PublicPictures = GUID.from_uuid(UUID('{B6EBFB86-6907-413C-9AF7-4FC2ABF07CC5}'))
    FOLDERID_PublicRingtones = GUID.from_uuid(UUID('{E555AB60-153B-4D17-9F04-A5FE99FC15EC}'))
    FOLDERID_PublicUserTiles = GUID.from_uuid(UUID('{0482af6c-08f1-4c34-8c90-e17ec98b1e17}'))
    FOLDERID_PublicVideos = GUID.from_uuid(UUID('{2400183A-6185-49FB-A2D8-4A392A602BA3}'))
    FOLDERID_QuickLaunch = GUID.from_uuid(UUID('{52a4f021-7b75-48a9-9f6b-4b87a210bc8f}'))
    FOLDERID_Recent = GUID.from_uuid(UUID('{AE50C081-EBD2-438A-8655-8A092E34987A}'))
    FOLDERID_RecordedTVLibrary = GUID.from_uuid(UUID('{1A6FDBA2-F42D-4358-A798-B74D745926C5}'))
    FOLDERID_RecycleBinFolder = GUID.from_uuid(UUID('{B7534046-3ECB-4C18-BE4E-64CD4CB7D6AC}'))
    FOLDERID_ResourceDir = GUID.from_uuid(UUID('{8AD10C31-2ADB-4296-A8F7-E4701232C972}'))
    FOLDERID_Ringtones = GUID.from_uuid(UUID('{C870044B-F49E-4126-A9C3-B52A1FF411E8}'))
    FOLDERID_RoamingAppData = GUID.from_uuid(UUID('{3EB685DB-65F9-4CF6-A03A-E3EF65729F3D}'))
    FOLDERID_RoamedTileImages = GUID.from_uuid(UUID('{AAA8D5A5-F1D6-4259-BAA8-78E7EF60835E}'))
    FOLDERID_RoamingTiles = GUID.from_uuid(UUID('{00BCFC5A-ED94-4e48-96A1-3F6217F21990}'))
    FOLDERID_SampleMusic = GUID.from_uuid(UUID('{B250C668-F57D-4EE1-A63C-290EE7D1AA1F}'))
    FOLDERID_SamplePictures = GUID.from_uuid(UUID('{C4900540-2379-4C75-844B-64E6FAF8716B}'))
    FOLDERID_SamplePlaylists = GUID.from_uuid(UUID('{15CA69B3-30EE-49C1-ACE1-6B5EC372AFB5}'))
    FOLDERID_SampleVideos = GUID.from_uuid(UUID('{859EAD94-2E85-48AD-A71A-0969CB56A6CD}'))
    FOLDERID_SavedGames = GUID.from_uuid(UUID('{4C5C32FF-BB9D-43b0-B5B4-2D72E54EAAA4}'))
    FOLDERID_SavedPictures = GUID.from_uuid(UUID('{3B193882-D3AD-4eab-965A-69829D1FB59F}'))
    FOLDERID_SavedPicturesLibrary = GUID.from_uuid(UUID('{E25B5812-BE88-4bd9-94B0-29233477B6C3}'))
    FOLDERID_SavedSearches = GUID.from_uuid(UUID('{7d1d3a04-debb-4115-95cf-2f29da2920da}'))
    FOLDERID_Screenshots = GUID.from_uuid(UUID('{b7bede81-df94-4682-a7d8-57a52620b86f}'))
    FOLDERID_SearchHistory = GUID.from_uuid(UUID('{0D4C3DB6-03A3-462F-A0E6-08924C41B5D4}'))
    FOLDERID_SearchTemplates = GUID.from_uuid(UUID('{7E636BFE-DFA9-4D5E-B456-D7B39851D8A9}'))
    FOLDERID_SendTo = GUID.from_uuid(UUID('{8983036C-27C0-404B-8F08-102D10DCFD74}'))
    FOLDERID_SidebarDefaultParts = GUID.from_uuid(UUID('{7B396E54-9EC5-4300-BE0A-2482EBAE1A26}'))
    FOLDERID_SidebarParts = GUID.from_uuid(UUID('{A75D362E-50FC-4fb7-AC2C-A8BEAA314493}'))
    FOLDERID_SkyDrive = GUID.from_uuid(UUID('{A52BBA46-E9E1-435f-B3D9-28DAA648C0F6}'))
    FOLDERID_SkyDriveCameraRoll = GUID.from_uuid(UUID('{767E6811-49CB-4273-87C2-20F355E1085B}'))
    FOLDERID_SkyDriveDocuments = GUID.from_uuid(UUID('{24D89E24-2F19-4534-9DDE-6A6671FBB8FE}'))
    FOLDERID_SkyDriveMusic = GUID.from_uuid(UUID('{C3F2459E-80D6-45DC-BFEF-1F769F2BE730}'))
    FOLDERID_SkyDrivePictures = GUID.from_uuid(UUID('{339719B5-8C47-4894-94C2-D8F77ADD44A6}'))
    FOLDERID_StartMenu = GUID.from_uuid(UUID('{625B53C3-AB48-4EC1-BA1F-A1EF4146FC19}'))
    FOLDERID_Startup = GUID.from_uuid(UUID('{B97D20BB-F46A-4C97-BA10-5E3608430854}'))
    FOLDERID_SyncManagerFolder = GUID.from_uuid(UUID('{43668BF8-C14E-49B2-97C9-747784D784B7}'))
    FOLDERID_SyncResultsFolder = GUID.from_uuid(UUID('{289a9a43-be44-4057-a41b-587a76d7e7f9}'))
    FOLDERID_SyncSetupFolder = GUID.from_uuid(UUID('{0F214138-B1D3-4a90-BBA9-27CBC0C5389A}'))
    FOLDERID_System = GUID.from_uuid(UUID('{1AC14E77-02E7-4E5D-B744-2EB1AE5198B7}'))
    FOLDERID_SystemX86 = GUID.from_uuid(UUID('{D65231B0-B2F1-4857-A4CE-A8E7C6EA7D27}'))
    FOLDERID_Templates = GUID.from_uuid(UUID('{A63293E8-664E-48DB-A079-DF759E0509F7}'))
    FOLDERID_UserPinned = GUID.from_uuid(UUID('{9E3995AB-1F9C-4F13-B827-48B24B6C7174}'))
    FOLDERID_UserProfiles = GUID.from_uuid(UUID('{0762D272-C50A-4BB0-A382-697DCD729B80}'))
    FOLDERID_UserProgramFiles = GUID.from_uuid(UUID('{5CD7AEE2-2219-4A67-B85D-6C9CE15660CB}'))
    FOLDERID_UserProgramFilesCommon = GUID.from_uuid(UUID('{BCBD3057-CA5C-4622-B42D-BC56DB0AE516}'))
    FOLDERID_UsersFiles = GUID.from_uuid(UUID('{f3ce0f7c-4901-4acc-8648-d5d44b04ef8f}'))
    FOLDERID_UsersLibraries = GUID.from_uuid(UUID('{A302545D-DEFF-464b-ABE8-61C8648D939B}'))
    FOLDERID_Videos = GUID.from_uuid(UUID('{18989B1D-99B5-455B-841C-AB7C74E4DDFC}'))
    FOLDERID_VideosLibrary = GUID.from_uuid(UUID('{491E922F-5643-4AF4-A7EB-4E7A138D8174}'))
    FOLDERID_Windows = GUID.from_uuid(UUID('{F38BF404-1D43-42F2-9305-67DE0B28FC23}'))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['run_forever', 'cancel_all_tasks']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"
from asyncio import coroutines
from asyncio import events
from asyncio import tasks


def run_forever(main, *, debug=False):
    
    """Execute the coroutine and all started tasks and return the result.

    This function runs the passed coroutine, taking care of
    managing the asyncio event loop and finalizing asynchronous
    generators.

    This function cannot be called when another asyncio event loop is
    running in the same thread.

    If debug is True, the event loop will be run in debug mode.

    This function always creates a new event loop and closes it at the end.
    It should be used as a main entry point for asyncio programs, and should
    ideally only be called once.

    Example:
        async def my_task():
            await asyncio.sleep(3)
            print('will be finished before loop ends unlike with the asyncio.run() call')


        async def main():
            asyncio.create_task(my_task())
            await asyncio.sleep(1)
            print('hello')


        run_forever(main())
    """
    if events._get_running_loop() is not None:
        raise RuntimeError(
            'run_forever() cannot be called from a running event loop')

    if not coroutines.iscoroutine(main):
        raise ValueError('a coroutine was expected, got {!r}'.format(main))

    loop = events.new_event_loop()
    try:
        events.set_event_loop(loop)
        loop.set_debug(debug)
        loop.create_task(main)
        return loop.run_forever()
    finally:
        try:
            cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            events.set_event_loop(None)
            loop.close()


def cancel_all_tasks(loop):
    to_cancel = tasks.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        tasks.gather(*to_cancel, loop=loop, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'unhandled exception during run_forever() shutdown',
                'exception': task.exception(),
                'task': task,
            })



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AnAppropriateValues', 'AllPossibleValues', 'AnAppropriateContainers', 'data_2_tuple', 'data_2_ordered', 'TreeTraversalType', 'TreeTraversal']
__all__ = ['AnAppropriateValues', 'AllPossibleValues', 'AnAppropriateContainers', 'TupledDict', 'Contaiters', 'data_2_tuple', 'data_2_ordered', 'TreeTraversalType', 'NodeType', 'EntityType', 'findout_entity_type', 'TreeTraversal', 'KeyMultiValueTreeTraversal', 'KeyValueTreeTraversal']


from functools import partial
from typing import Any, Hashable, Union, Dict, Set, List, Tuple, Optional, Callable
from enum import Enum
from collections import deque
from cengal.code_flow_control.smart_values import ValueExistence, ValueType
from cengal.data_containers.stack import StackItem, TreeStackItem, Stack, recursion_insureness

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


AnAppropriateValues = Union[str, bytes, int, float, None]
AllPossibleValues = Union['AnAppropriateContainers', AnAppropriateValues]
AnAppropriateContainers = Union[Dict[AnAppropriateValues, AllPossibleValues],
                                Set[AllPossibleValues],
                                List[AllPossibleValues],
                                Tuple[AllPossibleValues]]


class TupledDict:
    def __init__(self, data: Dict) -> None:
        self.dict: Dict = data
        self.items: Tuple[Tuple[Hashable, Any]] = tuple(data.items())


Contaiters = (TupledDict, dict, tuple, list, set, deque)


def data_2_tuple(data: AllPossibleValues) -> AllPossibleValues:
    if isinstance(data, Contaiters):
        if isinstance(data, dict):
            return tuple(data.values())
        elif isinstance(data, (tuple, list, deque)):
            return data
        elif isinstance(data, set):
            return tuple(data)
    else:
        return data


def data_2_ordered(data: AllPossibleValues) -> AllPossibleValues:
    if isinstance(data, Contaiters):
        if isinstance(data, dict):
            return TupledDict(data)
        elif isinstance(data, set):
            return tuple(data)
        else:
            return data
    else:
        return data


class TreeTraversalType(Enum):
    recursive = 0
    stack_based = 1
    recursive_with_stack_based_insureness = 2


class NodeType(Enum):
    dict = 0
    dict_item = 1
    sequence = 2


class EntityType(Enum):
    dict = 0
    dict_item = 1
    sequence = 2
    child = 3


def findout_entity_type(entity) -> EntityType:
    if isinstance(entity, Contaiters):
        if isinstance(entity, TupledDict):
            entity_type: EntityType = EntityType.dict
        else:
            entity_type = EntityType.sequence
    else:
        entity_type = EntityType.child
    
    return entity_type


class TreeTraversal:
    def __init__(self,
                 tree: AnAppropriateContainers=None,
                 on_node: Optional[Callable]=None,
                 on_child: Optional[Callable]=None,
                 on_switched_to_stack_based_implementation: Optional[Callable]=None
                 ):
        self._tree = tree  # type: AnAppropriateContainers

        self._callback_on_node = on_node  # type: Optional[Callable]
        self._callback_on_child = on_child  # type: Optional[Callable]
        self._on_switched_to_stack_based_implementation = \
            on_switched_to_stack_based_implementation  # type: Optional[Callable]

        self._last_tree_node = ValueExistence(False, None)
        self._last_tree_child = ValueExistence(False, None)
        self._empty_child = ValueExistence(False, None)

    def set_callback_on_node(self, functor: Optional[Callable]):
        self._callback_on_node = functor

    def set_callback_on_child(self, functor: Optional[Callable]):
        self._callback_on_child = functor

    def set_callback_on_both_node_and_child(self, functor: Optional[Callable]):
        self._callback_on_node = self._callback_on_child = functor

    def set_callback_on_switched_to_stack_based_implementation(self, functor: Optional[Callable]):
        self._on_switched_to_stack_based_implementation = functor

    @property
    def tree(self) -> AnAppropriateContainers:
        return self._tree

    @tree.setter
    def tree(self, tree: AnAppropriateContainers):
        self._tree = tree

    def _traversal_recursive(self):
        ordered_tree = data_2_ordered(self.tree)
        self._traversal_recursive_impl(ordered_tree)

    def _traversal_recursive_impl(self, entity: AnAppropriateContainers, deep: int = 0, index: Optional[int] = None):
        if isinstance(entity, Contaiters):
            last_tree_node_bak = self._last_tree_node
            if isinstance(entity, TupledDict):
                node_type: NodeType = NodeType.dict
            else:
                node_type = NodeType.sequence
            
            self._last_tree_node.value = ValueType(node_type, entity)
            if self._callback_on_node:
                self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

            new_deep = deep + 1
            if isinstance(entity, TupledDict):
                value_deep = new_deep + 1
                for index, node in enumerate(entity.items):
                    self._last_tree_node.value = ValueType(NodeType.dict_item, node)
                    if self._callback_on_node:
                        self._callback_on_node(new_deep, self._last_tree_node, self._empty_child, index)
                    
                    item = node[1]
                    self._traversal_recursive_impl(data_2_ordered(item), value_deep, 0)
            else:
                for index, item in enumerate(entity):
                    self._traversal_recursive_impl(data_2_ordered(item), new_deep, index)

            self._last_tree_node = last_tree_node_bak
        else:
            self._last_tree_child.value = entity
            if self._callback_on_child:
                self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)

    def _traversal_stack_based(self):
        index: Optional[int] = None
        deep: int = 0
        entity = data_2_ordered(self.tree)
        entity_type: EntityType = findout_entity_type(entity)
        stack = Stack(TreeStackItem((entity_type, entity, deep, index)))
        while stack:
            top = stack.top()
            entity_type, entity, deep, index = top.node
            if EntityType.child == entity_type:
                self._last_tree_child.value = entity
                if self._callback_on_child:
                    self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)
                
                stack.pop()
            elif EntityType.sequence == entity_type:
                self._last_tree_node.value = ValueType(NodeType.sequence, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity) > child_index:
                    top.child = child_index
                    child = data_2_ordered(entity[child_index])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict_item == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict_item, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

                    top.child = child_index
                    child = data_2_ordered(entity[1])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict, entity.dict)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity.items) > child_index:
                    top.child = child_index
                    child = entity.items[child_index]
                    stack.push(TreeStackItem((EntityType.dict_item, child, deep + 1, child_index)))
                else:
                    stack.pop()

    def __call__(self, traversal_type: TreeTraversalType=TreeTraversalType.recursive_with_stack_based_insureness):
        if TreeTraversalType.recursive == traversal_type:
            self._traversal_recursive()
        elif TreeTraversalType.stack_based == traversal_type:
            self._traversal_stack_based()
        else:
            recursion_insureness(self._traversal_recursive,
                                 self._traversal_stack_based,
                                 self._on_switched_to_stack_based_implementation)


class KeyMultiValueTreeTraversal:
    def __init__(self,
                 tree: AnAppropriateContainers=None,
                 on_node: Optional[Callable]=None,
                 on_child: Optional[Callable]=None,
                 on_switched_to_stack_based_implementation: Optional[Callable]=None
                 ):
        self._tree = tree  # type: AnAppropriateContainers

        self._callback_on_node = on_node  # type: Optional[Callable]
        self._callback_on_child = on_child  # type: Optional[Callable]
        self._on_switched_to_stack_based_implementation = \
            on_switched_to_stack_based_implementation  # type: Optional[Callable]

        self._last_tree_node = ValueExistence(False, None)
        self._last_tree_child = ValueExistence(False, None)
        self._empty_child = ValueExistence(False, None)

    def set_callback_on_node(self, functor: Optional[Callable]):
        self._callback_on_node = functor

    def set_callback_on_child(self, functor: Optional[Callable]):
        self._callback_on_child = functor

    def set_callback_on_both_node_and_child(self, functor: Optional[Callable]):
        self._callback_on_node = self._callback_on_child = functor

    def set_callback_on_switched_to_stack_based_implementation(self, functor: Optional[Callable]):
        self._on_switched_to_stack_based_implementation = functor

    @property
    def tree(self) -> AnAppropriateContainers:
        return self._tree

    @tree.setter
    def tree(self, tree: AnAppropriateContainers):
        self._tree = tree

    def _traversal_recursive(self, initial_key: Hashable):
        index: Optional[int] = None
        deep: int = 0
        self._last_tree_child.value = initial_key
        if self._callback_on_child:
            self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)

        try:
            entity = self.tree[initial_key]
        except KeyError:
            return

        self._last_tree_node.value = initial_key
        if self._callback_on_node:
            self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

        entity = data_2_ordered(entity)
        self._traversal_recursive_impl(entity, deep, index)

    def _traversal_recursive_impl(self, entity: AnAppropriateContainers, deep: int = 0, index: Optional[int] = None):
        if isinstance(entity, Contaiters):
            last_tree_node_bak = self._last_tree_node
            new_deep = deep + 1
            for index, item in enumerate(entity):
                self._traversal_recursive_impl(item, new_deep, index)
                self._last_tree_node = last_tree_node_bak
        else:
            key = entity
            index: Optional[int] = None
            deep: int = 0
            self._last_tree_child.value = key
            if self._callback_on_child:
                self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)
            
            try:
                entity = self.tree[key]
            except KeyError:
                return

            self._last_tree_node.value = key
            if self._callback_on_node:
                self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

            entity = data_2_ordered(entity)
            self._traversal_recursive_impl(entity, deep, index)

    def _traversal_stack_based(self, initial_key: Hashable):
        raise NotImplementedError
        index: Optional[int] = None
        deep: int = 0
        key: Hashable = initial_key
        entity = data_2_ordered(self.tree)
        entity_type: EntityType = findout_entity_type(entity)
        stack = Stack(TreeStackItem((entity_type, entity, key, deep, index)))
        while stack:
            top = stack.top()
            entity_type, entity, key, deep, index = top.node
            if EntityType.child == entity_type:
                self._last_tree_child.value = entity
                if self._callback_on_child:
                    self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)
                
                stack.pop()
            elif EntityType.sequence == entity_type:
                self._last_tree_node.value = ValueType(NodeType.sequence, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity) > child_index:
                    top.child = child_index
                    child = data_2_ordered(entity[child_index])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict_item == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict_item, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

                    top.child = child_index
                    child = data_2_ordered(entity[1])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict, entity.dict)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity.items) > child_index:
                    top.child = child_index
                    child = entity.items[child_index]
                    stack.push(TreeStackItem((EntityType.dict_item, child, deep + 1, child_index)))
                else:
                    stack.pop()

    def __call__(self, initial_key: Hashable, traversal_type: TreeTraversalType=TreeTraversalType.recursive_with_stack_based_insureness):
        if TreeTraversalType.recursive == traversal_type:
            self._traversal_recursive(initial_key)
        elif TreeTraversalType.stack_based == traversal_type:
            self._traversal_stack_based(initial_key)
        else:
            recursive = partial(self._traversal_recursive, initial_key)
            stack_based = partial(self._traversal_stack_based, initial_key)
            recursion_insureness(recursive,
                                 stack_based,
                                 self._on_switched_to_stack_based_implementation)


class KeyValueTreeTraversal:
    def __init__(self,
                 tree: AnAppropriateContainers=None,
                 on_node: Optional[Callable]=None,
                 on_child: Optional[Callable]=None,
                 on_switched_to_stack_based_implementation: Optional[Callable]=None
                 ):
        self._tree = tree  # type: AnAppropriateContainers

        self._callback_on_node = on_node  # type: Optional[Callable]
        self._callback_on_child = on_child  # type: Optional[Callable]
        self._on_switched_to_stack_based_implementation = \
            on_switched_to_stack_based_implementation  # type: Optional[Callable]

        self._last_tree_node = ValueExistence(False, None)
        self._last_tree_child = ValueExistence(False, None)
        self._empty_child = ValueExistence(False, None)

    def set_callback_on_node(self, functor: Optional[Callable]):
        self._callback_on_node = functor

    def set_callback_on_child(self, functor: Optional[Callable]):
        self._callback_on_child = functor

    def set_callback_on_both_node_and_child(self, functor: Optional[Callable]):
        self._callback_on_node = self._callback_on_child = functor

    def set_callback_on_switched_to_stack_based_implementation(self, functor: Optional[Callable]):
        self._on_switched_to_stack_based_implementation = functor

    @property
    def tree(self) -> AnAppropriateContainers:
        return self._tree

    @tree.setter
    def tree(self, tree: AnAppropriateContainers):
        self._tree = tree

    def _traversal_recursive(self, initial_key: Hashable):
        index: Optional[int] = None
        deep: int = 0
        self._last_tree_child.value = initial_key
        if self._callback_on_child:
            self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)

        try:
            entity = self.tree[initial_key]
        except KeyError:
            return

        self._last_tree_node.value = initial_key
        if self._callback_on_node:
            self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

        self._traversal_recursive_impl(entity, deep + 1, index)

    def _traversal_recursive_impl(self, entity: AnAppropriateContainers, deep: int = 0, index: Optional[int] = None):
        key = entity
        index: Optional[int] = None
        deep: int = 0
        self._last_tree_child.value = key
        if self._callback_on_child:
            self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)
        
        try:
            entity = self.tree[key]
        except KeyError:
            return

        self._last_tree_node.value = key
        if self._callback_on_node:
            self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

        self._traversal_recursive_impl(entity, deep + 1, index)

    def _traversal_stack_based(self, initial_key: Hashable):
        raise NotImplementedError
        index: Optional[int] = None
        deep: int = 0
        key: Hashable = initial_key
        entity = data_2_ordered(self.tree)
        entity_type: EntityType = findout_entity_type(entity)
        stack = Stack(TreeStackItem((entity_type, entity, key, deep, index)))
        while stack:
            top = stack.top()
            entity_type, entity, key, deep, index = top.node
            if EntityType.child == entity_type:
                self._last_tree_child.value = entity
                if self._callback_on_child:
                    self._callback_on_child(deep, self._last_tree_node, self._last_tree_child, index)
                
                stack.pop()
            elif EntityType.sequence == entity_type:
                self._last_tree_node.value = ValueType(NodeType.sequence, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity) > child_index:
                    top.child = child_index
                    child = data_2_ordered(entity[child_index])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict_item == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict_item, entity)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)

                    top.child = child_index
                    child = data_2_ordered(entity[1])
                    child_type = findout_entity_type(child)
                    stack.push(TreeStackItem((child_type, child, deep + 1, child_index)))
                else:
                    stack.pop()
            elif EntityType.dict == entity_type:
                self._last_tree_node.value = ValueType(NodeType.dict, entity.dict)
                if top.child is None:
                    child_index = 0
                    if self._callback_on_node:
                        self._callback_on_node(deep, self._last_tree_node, self._empty_child, index)
                else:
                    child_index = top.child + 1
                
                if len(entity.items) > child_index:
                    top.child = child_index
                    child = entity.items[child_index]
                    stack.push(TreeStackItem((EntityType.dict_item, child, deep + 1, child_index)))
                else:
                    stack.pop()

    def __call__(self, initial_key: Hashable, traversal_type: TreeTraversalType=TreeTraversalType.recursive_with_stack_based_insureness):
        if TreeTraversalType.recursive == traversal_type:
            self._traversal_recursive(initial_key)
        elif TreeTraversalType.stack_based == traversal_type:
            self._traversal_stack_based(initial_key)
        else:
            recursive = partial(self._traversal_recursive, initial_key)
            stack_based = partial(self._traversal_stack_based, initial_key)
            recursion_insureness(recursive,
                                 stack_based,
                                 self._on_switched_to_stack_based_implementation)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import gc
from cengal.code_flow_control.gc import DisableGC
from contextlib import contextmanager
from cengal.code_flow_control.smart_values.versions.v_2 import ValueExistence
from cengal.parallel_execution.coroutines.coro_standard_services.lazy_print.versions.v_0.lazy_print import lprint
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.time_management.load_best_timer import process_time
from cengal.time_management.repeat_for_a_time import Tracer, ClockType
from cengal.math.numbers import RationalNumber
from cengal.introspection.inspect import is_async
from cengal.introspection.inspect import func_name, entity_properties
from cengal.code_inspection.auto_line_tracer import alt, LineType, OutputFields, AutoLineTracer

from typing import Union, Callable, Awaitable, List, Optional, Type, Set

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class PerformanceTestResult(Exception):
    def __init__(self, result):
        super(PerformanceTestResult, self).__init__()
        self.result = result


@contextmanager
def test_run_time(test_name: str, number_of_iterations: int, throw_result: bool=False, throw_result_anyway: bool=True, ignore_index=False):
    index = ValueExistence(True, copy.copy(number_of_iterations))
    start_time = perf_counter()
    exception_occures = False
    try:
        yield index
    except:
        if not throw_result_anyway:
            exception_occures = True
        raise
    finally:
        if not ignore_index:
            number_of_iterations -= index.value
        
        end_time = perf_counter()
        result_time = end_time - start_time
        if result_time > 0:
            text_result = f'>>> "{test_name}"\n\tIt was used {result_time} seconds to process {number_of_iterations} iterations.\n\tThere is {number_of_iterations / result_time} iterations per second\n'
        else:
            text_result = f'>>> "{test_name}"\n\tIt was used {result_time} seconds to process {number_of_iterations} iterations.\n'

        lprint(text_result)

        if (not exception_occures) and throw_result:
            result_data = dict()
            result_data['test_name'] = test_name
            result_data['result_time'] = result_time
            if result_time > 0:
                result_data['iterations_per_time_unit'] = number_of_iterations / result_time
            else:
                result_data['iterations_per_time_unit'] = None
            raise PerformanceTestResult(result_data)


def measure_time(test_name: str = str()):
    return test_run_time(test_name, 1, ignore_index=True)


def test_function_run_time(
        test_name: str = None,
        iterations_qnt: int = None,
        throw_result: bool = None,
    ):
    """
    Use 'performance_test_lib__iterations_qnt=1000000' parameter to pass number of iterations
    :param testable_function: function
    :return:
    """
    def wrapper(testable_function):
        if is_async(testable_function):
            async def sub_wrapper(*args, **kwargs):
                test_name_ = '' if test_name is None else test_name
                test_full_name = '{}: {}'.format(str(testable_function), test_name_)
                number_of_iterations = 1 if iterations_qnt is None else iterations_qnt
                throw_result_ = False if throw_result is None else throw_result
                with test_run_time(test_full_name, number_of_iterations, throw_result_) as index:
                    while index.value > 0:
                        await testable_function(*args, **kwargs)
                        index.value -= 1
            
            return sub_wrapper
        else:
            def sub_wrapper(*args, **kwargs):
                test_name_ = '' if test_name is None else test_name
                test_full_name = '{}: {}'.format(str(testable_function), test_name_)
                number_of_iterations = 1 if iterations_qnt is None else iterations_qnt
                throw_result_ = False if throw_result is None else throw_result
                with test_run_time(test_full_name, number_of_iterations, throw_result_) as index:
                    while index.value > 0:
                        testable_function(*args, **kwargs)
                        index.value -= 1
            
            return sub_wrapper
    return wrapper


def process_performance_test_results(tracer: Tracer, test_name: str, throw_result: bool=False):
    number_of_iterations = tracer.iterations_made
    result_time = tracer.time_spent
    iterations_per_time_unit = tracer.iter_per_time_unit
    print('>>> "{}"'.format(test_name))
    print('\t' + 'It was used', result_time, 'seconds to process', number_of_iterations, 'iterations.')
    print('\t' + 'There is', iterations_per_time_unit, 'iterations per second')

    if throw_result:
        result_data = (test_name, result_time, iterations_per_time_unit)
        raise PerformanceTestResult(result_data)


@contextmanager
def test_performance(test_name: str, run_time: float, throw_result: bool=False, clock_type=ClockType.perf_counter):
    tracer = Tracer(run_time, clock_type)
    try:
        yield tracer
    except:
        raise
    finally:
        process_performance_test_results(tracer, test_name, throw_result)


def test_function_performance(
        test_name: str = None,
        run_time: RationalNumber = None,
        throw_result: bool = None,
        clock_type: ClockType = None,
    ):
    """
    Use 'performance_test_lib__run_time=1.5' parameter to pass number of seconds to test
    :param testable_function: function
    :return:
    """
    def wrapper(testable_function):
        if is_async(testable_function):
            async def sub_wrapper(*args, **kwargs):
                test_name_ = '' if test_name is None else test_name
                full_test_name = '{}: {}'.format(str(testable_function), test_name_)
                run_time_ = 1.0 if run_time is None else run_time
                throw_result_ = False if throw_result is None else throw_result
                clock_type_ = ClockType.perf_counter if clock_type is None else clock_type
                with test_performance(full_test_name, run_time_, throw_result_, clock_type_) as tracer:
                    while tracer.iter():
                        await testable_function(*args, **kwargs)
            
            return sub_wrapper
        else:
            def sub_wrapper(*args, **kwargs):
                test_name_ = '' if test_name is None else test_name
                full_test_name = '{}: {}'.format(str(testable_function), test_name_)
                run_time_ = 1.0 if run_time is None else run_time
                throw_result_ = False if throw_result is None else throw_result
                clock_type_ = ClockType.perf_counter if clock_type is None else clock_type
                with test_performance(full_test_name, run_time_, throw_result_, clock_type_) as tracer:
                    while tracer.iter():
                        testable_function(*args, **kwargs)
            
            return sub_wrapper
    return wrapper


class PrecisePerformanceTestTracer(Tracer):
    """
    Precise tracer.
    At first you need to use it as a usual Tracer. After tracing was done - use it as a fast `for i in range(...)` block

    Example of use:

        tr = PrecisePerformanceTestTracer(10.0)
        while tr.iter():
            i = '456'
            k = int('1243' + i)

        with tr as fast_iter:
            for i in fast_iter:
                i = '456'
                k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    """

    def __init__(self,
                 run_time: float,
                 clock_type: ClockType=ClockType.perf_counter,
                 suppress_exceptions: bool=False,
                 turn_off_gc: bool=False
                 ):
        super().__init__(run_time, clock_type)
        self.suppress_exceptions = suppress_exceptions
        self.turn_off_gc = turn_off_gc
        self.gc_was_enabled = None

    def __enter__(self):
        self._relevant_start_time = self._start_time = self._relevant_stop_time = self._end_time = self._clock()
        self._relevant_number_of_iterations_at_start = 0
        if self.turn_off_gc:
            self.gc_was_enabled = gc.isenabled()
            gc.disable()
        
        return range(self._last_tracked_number_of_iterations)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._relevant_stop_time = self._end_time = self._clock()
        if self.turn_off_gc and self.gc_was_enabled:
            gc.enable()
        
        if self.suppress_exceptions:
            return True


class MeasureTime:
    def __init__(self, name: str=None, iterations: int = 1, do_print: Union[bool, Callable] = True, raise_exceptions: bool = True):
        self.name: str = name
        self.iterations: int = 1 if iterations < 1 else iterations
        self.do_print: Union[bool, Callable] = do_print
        self.raise_exceptions: bool = raise_exceptions
        self.start_time = None
        self.stop_time = None
        self.time_spent = None
        self.exc_type = None
        self.exc_value = None
        self.exc_tb = None

    def __enter__(self):
        self.start_time = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exc_type = exc_type
        self.exc_value = exc_val
        self.exc_tb = exc_tb

        self.stop_time = perf_counter()
        self.time_spent = self.stop_time - self.start_time
        if self.do_print:
            if isinstance(self.do_print, bool):
                if self.name is not None:
                    print(f'>>> "{self.name}"')
                else:
                    print('>>>')

                if self.exc_type is not None:
                    print(f'\t Exception: {self.exc_type}')
                    print(f'\t Exception value: {self.exc_value}')
                    print(f'\t Exception traceback: {self.exc_tb}')
                
                if self.iterations > 1:
                    print(f'\t It was used {self.time_spent} secondss to process {self.iterations} iterations')
                else:
                    print(f'\t It was used {self.time_spent} seconds')
                
                if self.time_spent:
                    print(f'\t There is {self.iterations / self.time_spent} iterations/seconds')
                    print()
            else:
                self.do_print(self)
        
        return not self.raise_exceptions


class MeasureTimeTraceLine:
    def __init__(self, name: str=None, iterations: int = 1, raise_exceptions: bool = True, measuring_class: Type = MeasureTime, 
                 line_type: LineType = LineType.current_line, line_num: Optional[int] = None, output_fields: Optional[Set[OutputFields]] = None, 
                 do_print: Union[bool, Callable] = True, auto_line_tracer: AutoLineTracer = None, depth: int = 1):
        self.measuring_class: Type = measuring_class
        self.measuring_obj: MeasureTime = self.measuring_class(name, iterations, self.printer if do_print is True else do_print, raise_exceptions)
        self.depth: int = depth
        self.output_fields: Set[OutputFields] = {
            OutputFields.trace_name, 
            OutputFields.file_name,
            OutputFields.line,
            OutputFields.func_name,
            OutputFields.code_line,
            OutputFields.new_line_after_end,
        } if output_fields is None else output_fields
        self.line_type: LineType = line_type
        self.line_num: Optional[int] = line_num
        self.auto_line_tracer: AutoLineTracer = alt if auto_line_tracer is None else auto_line_tracer
    
    def printer(self, measuring_obj: MeasureTime) -> None:
        print('='*40)
        if self.measuring_obj.name is not None:
            print(f'>>> "{measuring_obj.name}"')

        self.auto_line_tracer.pc('', line_type=self.line_type, line_num=self.line_num, output_fields=self.output_fields, depth=self.depth + 3)

        if measuring_obj.exc_type is not None:
            print(f'\t Exception: {measuring_obj.exc_type}')
            print(f'\t Exception value: {measuring_obj.exc_value}')
            print(f'\t Exception traceback: {measuring_obj.exc_tb}')
        
        if measuring_obj.iterations > 1:
            print(f'It was used {measuring_obj.time_spent} secondss to process {measuring_obj.iterations} iterations')
        else:
            print(f'It was used {measuring_obj.time_spent} seconds')
        
        if measuring_obj.time_spent:
            print(f'There is {measuring_obj.iterations / measuring_obj.time_spent} iterations/seconds')
        
        print()

    def __enter__(self):
        return self.measuring_obj.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.measuring_obj.__exit__(exc_type, exc_val, exc_tb)


class MeasureProcessTime:
    def __init__(self, name: str=None, iterations: int = 1, do_print: Union[bool, Callable] = False, raise_exceptions: bool = True, depth: int = 1):
        self.name: str = name
        self.iterations: int = 1 if iterations < 1 else iterations
        self.do_print: Union[bool, Callable] = do_print
        self.raise_exceptions: bool = raise_exceptions
        self.depth: int = depth
        self.start_time = None
        self.stop_time = None
        self.time_spent = None
        self.exc_type = None
        self.exc_value = None
        self.exc_tb = None

    def __enter__(self):
        self.start_time = process_time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exc_type = exc_type
        self.exc_value = exc_val
        self.exc_tb = exc_tb

        self.stop_time = process_time()
        self.time_spent = self.stop_time - self.start_time
        if self.do_print:
            if isinstance(self.do_print, bool):
                if self.name is not None:
                    print(f'>>> "{self.name}"')
                else:
                    print('>>>')

                if self.exc_type is not None:
                    print(f'\t Exception: {self.exc_type}')
                    print(f'\t Exception value: {self.exc_value}')
                    print(f'\t Exception traceback: {self.exc_tb}')
                
                if self.iterations > 1:
                    print(f'\t It was used {self.time_spent} secondss to process {self.iterations} iterations')
                else:
                    print(f'\t It was used {self.time_spent} seconds')
                
                if self.time_spent:
                    print(f'\t There is {self.iterations / self.time_spent} iterations/seconds')
                    print()
            else:
                self.do_print(self)
        
        return not self.raise_exceptions


def measure_func_performance(func: Callable, 
                    run_time: float,
                    name: str=None, 
                    do_print: Union[bool, Callable] = False, 
                    clock_type: ClockType=ClockType.perf_counter,
                    suppress_exceptions: bool=False,
                    turn_off_gc: bool=False
                   ):
    tr = PrecisePerformanceTestTracer(run_time, clock_type, suppress_exceptions, turn_off_gc)
    try:
        if turn_off_gc:
            gc_was_enabled = gc.isenabled()
            gc.disable()
        
        while tr.iter():
            func()

        with tr as fast_iter:
            for i in fast_iter:
                func()
    finally:
        if turn_off_gc and gc_was_enabled:
            gc.enable()

    if do_print:
        if isinstance(do_print, bool):
            if name is not None:
                print(f'>>> "{name}": {func_name(func)}()')
            else:
                print(f'>>> {func_name(func)}()')

            print(f'\t It was used {tr.time_spent} seconds to make {tr.iterations_made} iterations. Performance: {tr.iter_per_time_unit} iterations/seconds')
        else:
            do_print(func, run_time, name, clock_type, tr)


async def measure_afunc_performance(afunc: Awaitable, 
                    run_time: float,
                    name: str=None, 
                    do_print: Union[bool, Callable] = False, 
                    clock_type: ClockType=ClockType.perf_counter,
                    suppress_exceptions: bool=False,
                    turn_off_gc: bool=False
                   ):
    tr = PrecisePerformanceTestTracer(run_time, clock_type, suppress_exceptions, turn_off_gc)
    try:
        if turn_off_gc:
            gc_was_enabled = gc.isenabled()
            gc.disable()
        
        while tr.iter():
            await afunc

        with tr as fast_iter:
            for i in fast_iter:
                await afunc
    finally:
        if turn_off_gc and gc_was_enabled:
            gc.enable()

    if do_print:
        if isinstance(do_print, bool):
            if name is not None:
                print(f'>>> "{name}": {func_name(afunc)}()')
            else:
                print(f'>>> {func_name(afunc)}()')

            print(f'\t It was used {tr.time_spent} seconds to make {tr.iterations_made} iterations. Performance: {tr.iter_per_time_unit} iterations/seconds')
        else:
            do_print(afunc, run_time, name, clock_type, tr)


def measure_func_isolated_performance(func: Callable, 
                    run_time: float,
                    name: str=None, 
                    do_print: Union[bool, Callable] = False, 
                    clock_type: ClockType=ClockType.perf_counter,
                    suppress_exceptions: bool=False,
                    turn_off_gc: bool=False
                   ):
    """Measure functions best/top/isolated performance. Hopefully isolated from an interference with other threads executed on the same CPU core

    Args:
        func (Callable): _description_
        run_time (float): _description_
        name (str, optional): _description_. Defaults to None.
        do_print (Union[bool, Callable], optional): _description_. Defaults to False.
        clock_type (ClockType, optional): _description_. Defaults to ClockType.perf_counter.
        suppress_exceptions (bool, optional): _description_. Defaults to False.
        turn_off_gc (bool, optional): _description_. Defaults to False.
    """    
    tr = PrecisePerformanceTestTracer(run_time, clock_type, suppress_exceptions, turn_off_gc)
    measurements: List[float] = list()
    try:
        if turn_off_gc:
            gc_was_enabled = gc.isenabled()
            gc.disable()
        
        while tr.iter():
            func()

        with tr as fast_iter:
            for i in fast_iter:
                start_time = perf_counter()
                func()
                measurements.append(perf_counter() - start_time)
    finally:
        if turn_off_gc and gc_was_enabled:
            gc.enable()

    if do_print:
        best_measurement: float = min(measurements)
        best_performance: Optional[float] = (1 / best_measurement) if best_measurement else None
        if isinstance(do_print, bool):
            if name is not None:
                print(f'>>> "{name}": {func_name(func)}()')
            else:
                print(f'>>> {func_name(func)}()')

            print(f'\t It was used {tr.time_spent} seconds to make {tr.iterations_made} iterations. Performance: {tr.iter_per_time_unit} iterations/seconds')
            print(f'\t Isolated run time: {best_measurement} seconds; Isolated performance: {best_performance} iterations/seconds')
        else:
            do_print(func, run_time, name, clock_type, tr, best_measurement, best_performance)


async def measure_afunc_isolated_performance(afunc: Awaitable, 
                    run_time: float,
                    name: str=None, 
                    do_print: Union[bool, Callable] = False, 
                    clock_type: ClockType=ClockType.perf_counter,
                    suppress_exceptions: bool=False,
                    turn_off_gc: bool=False
                   ):
    """Measure functions best/top/isolated performance. Hopefully isolated from an interference with other threads executed on the same CPU core

    Args:
        afunc (Awaitable): _description_
        run_time (float): _description_
        name (str, optional): _description_. Defaults to None.
        do_print (Union[bool, Callable], optional): _description_. Defaults to False.
        clock_type (ClockType, optional): _description_. Defaults to ClockType.perf_counter.
        suppress_exceptions (bool, optional): _description_. Defaults to False.
        turn_off_gc (bool, optional): _description_. Defaults to False.
    """    
    tr = PrecisePerformanceTestTracer(run_time, clock_type, suppress_exceptions, turn_off_gc)
    measurements: List[float] = list()
    try:
        if turn_off_gc:
            gc_was_enabled = gc.isenabled()
            gc.disable()
        
        while tr.iter():
            await afunc

        with tr as fast_iter:
            for i in fast_iter:
                start_time = perf_counter()
                await afunc
                measurements.append(perf_counter() - start_time)
    finally:
        if turn_off_gc and gc_was_enabled:
            gc.enable()

    if do_print:
        best_measurement: float = min(measurements)
        best_performance: Optional[float] = (1 / best_measurement) if best_measurement else None
        if isinstance(do_print, bool):
            if name is not None:
                print(f'>>> "{name}": {func_name(afunc)}()')
            else:
                print(f'>>> {func_name(afunc)}()')

            print(f'\t It was used {tr.time_spent} seconds to make {tr.iterations_made} iterations. Performance: {tr.iter_per_time_unit} iterations/seconds')
            print(f'\t Isolated run time: {best_measurement} seconds; Isolated performance: {best_performance} iterations/seconds')
        else:
            do_print(afunc, run_time, name, clock_type, tr, best_measurement, best_performance)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['TimerHandler', 'TimeInSeconds', 'TimerRequest', 'Timer']


from time import perf_counter
from cengal.math.numbers import RationalNumber
from typing import Any, Callable, Optional, Set

TimerHandler = Callable
TimeInSeconds = RationalNumber


class TimerRequest:
    def __init__(self, timer_handler: TimerHandler, desired_time: TimeInSeconds):
        self.timer_handler = timer_handler
        self.requested_time = desired_time
        self.start_time = perf_counter()
        self.real_end_time = None
        self.processed = False

    def __call__(self) -> bool:
        if not self.processed:
            current_time = perf_counter()
            if self.start_time > current_time:
                self.start_time = current_time
            time_delta = current_time - self.start_time
            if time_delta >= self.requested_time:
                self.timer_handler()
                self.processed = True
        return self.processed

    def time_left(self) -> Optional[TimeInSeconds]:
        if self.processed:
            return None

        current_time = perf_counter()
        if self.start_time > current_time:
            self.start_time = current_time
        time_delta = current_time - self.start_time
        if time_delta >= self.requested_time:
            return 0
        else:
            return self.requested_time - time_delta


class Timer:
    # TODO: improve algorithm using `cengal\time_management\repeat_for_a_time` module
    def __init__(self):
        self.requests: Set[TimerRequest] = set()  # type: Set[TimerRequest]

    def register(self, timer_handler: TimerHandler, desired_time: TimeInSeconds) -> TimerRequest:
        request = TimerRequest(timer_handler, desired_time)
        self.requests.add(request)
        return request

    def discard(self, timer_request: TimerRequest):
        try:
            self.requests.remove(timer_request)
            return True
        except KeyError:
            return False

    def __call__(self):
        # processed_requests = set()
        # current_requests = self.requests
        # for request in current_requests:
        #     if request():
        #         processed_requests.add(request)

        # if processed_requests:
        #     self.requests -= processed_requests

        # requests_buff = self.requests
        # self.requests = type(requests_buff)()
        # for request in requests_buff:
        #     if not request():
        #         self.requests.add(request)

        requests_buff = self.requests
        self.requests = set()
        processed_requests = set()
        try:
            for request in requests_buff:
                request_result: bool = True
                try:
                    request_result = request()
                finally:
                    if request_result:
                        processed_requests.add(request)
        finally:
            if processed_requests:
                requests_buff -= processed_requests

            if self.requests:
                requests_buff.update(self.requests)

            self.requests = requests_buff

    def nearest_event(self) -> Optional[TimeInSeconds]:
        nearest_event_time = None
        for request in self.requests:
            time_left = request.time_left()
            if time_left is None:
                continue

            if 0 == time_left:
                return 0

            if (nearest_event_time is None) or (nearest_event_time > time_left):
                nearest_event_time = time_left

        return nearest_event_time



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.data_manipulation.remote_objects.versions.v_0.remote_objects import *
from cengal.code_flow_control.smart_values import ResultExistence
from cengal.performance_test_lib import MeasureTime
from cengal.introspection.inspect import cen
from collections import deque
from dataclasses import dataclass
from unittest import TestCase, main
from pickle import loads as pickle_loads, dumps as pickle_dumps
from typing import Any, Dict, Optional, Callable, Set, Type, Tuple, List, FrozenSet, Deque


class SimpleClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        return False


@dataclass
class DataClass:
    x: int
    y: 'Any'
    z: int = 5

    def add_one(self):
        return self.x + 1


class TestRemoteObjectsManager(TestCase):
    def setUp(self):
        "Hook method for setting up the test fixture before exercising it."
        self.rom_src: RemoteObjectsManager = RemoteObjectsManager(
            on_new_class_handler=self.on_new_class,
            on_new_obj_info_handler=self.on_new_object,
        )
        self.transport: Deque = deque()
        self.rom_dest: RemoteObjectsManager = RemoteObjectsManager(
            on_new_class_handler=self.on_new_class,
            on_new_obj_info_handler=self.on_new_object,
        )

        self.rom_src_pickle: RemoteObjectsManager = RemoteObjectsManager(
            on_new_class_handler=self.on_new_class_pickle,
            on_new_obj_info_handler=self.on_new_object_pickle,
            serializable_data_types=set(),
        )
        self.transport: Deque = deque()
        self.rom_dest_pickle: RemoteObjectsManager = RemoteObjectsManager(
            on_new_class_handler=self.on_new_class_pickle,
            on_new_obj_info_handler=self.on_new_object_pickle,
            serializable_data_types=set(),
        )

    def on_new_class(
            self,
            class_info,
            obj_type,
            class_id,
            class_name,
            module_importable_str,
            owning_names_path,
            module_importable_str_and_owning_names_path
        ):
        self.transport.append(
            (
                'new_class',
                class_info,
            )
        )
    
    def on_new_object(
            self,
            adjusted_object_info,
            obj,
            object_id,
            type_id,
            serializable,
            known_container,
            known_data,
            class_id,
            is_new_class,
            new_obj_slots,
            new_obj_dict,
            obj_mapping,
            obj_sequence,
            obj_set,
    ):
        self.transport.append(
            (
                'new_obj',
                adjusted_object_info,
            )
        )

    def on_new_class_pickle(
            self,
            class_info,
            obj_type,
            class_id,
            class_name,
            module_importable_str,
            owning_names_path,
            module_importable_str_and_owning_names_path
        ):
        self.transport.append(
            pickle_dumps((
                'new_class',
                class_info,
            ))
        )
    
    def on_new_object_pickle(
            self,
            adjusted_object_info,
            obj,
            object_id,
            type_id,
            serializable,
            known_container,
            known_data,
            class_id,
            is_new_class,
            new_obj_slots,
            new_obj_dict,
            obj_mapping,
            obj_sequence,
            obj_set,
    ):
        self.transport.append(
            pickle_dumps((
                'new_obj',
                adjusted_object_info,
            ))
        )

    def tearDown(self):
        "Hook method for deconstructing the test fixture after testing it."
        self.rom_src = None
        self.rom_dest = None
    
    def test_transmit_serializable(self):
        obj = {'a': 1, 'b': 2}
        with MeasureTime(f'{cen()} - codec'):
            with MeasureTime(f'{cen()} - serialization'):
                obj_id_result: ResultExistence[int] = self.rom_src.serialize(obj)
            
            self.assertTrue(obj_id_result)
            self.assertEqual(len(self.transport), 5)
            with MeasureTime(f'{cen()} - deserialization'):
                for item in self.transport:
                    item_type, item_info = item
                    self.assertIn(item_type, ('new_class', 'new_obj'))
                    if 'new_class' == item_type:
                        transmited_class_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_obj(item_info)
                        self.assertTrue(transmited_class_result)
                    elif 'new_obj' == item_type:
                        transmited_obj_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_obj(item_info)
                        self.assertTrue(transmited_obj_result)
        
        received_object_id, received_obj = transmited_obj_result.value
        self.assertDictEqual(obj, received_obj)
    
    def test_transmit_simple_class(self):
        obj = ({'a': 1, 'b': 2}, SimpleClass(1, 2))
        with MeasureTime(f'{cen()} - codec'):
            with MeasureTime(f'{cen()} - serialization'):
                obj_id_result: ResultExistence[int] = self.rom_src.serialize(obj)

            self.assertTrue(obj_id_result)
            self.assertEqual(len(self.transport), 10)
            with MeasureTime(f'{cen()} - deserialization'):
                for item in self.transport:
                    item_type, item_info = item
                    self.assertIn(item_type, ('new_class', 'new_obj'))
                    if 'new_class' == item_type:
                        transmited_class_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_class(item_info)
                        self.assertTrue(transmited_class_result)
                    elif 'new_obj' == item_type:
                        transmited_obj_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_obj(item_info)
                        self.assertTrue(transmited_obj_result)
        
        received_object_id, received_obj = transmited_obj_result.value
        dict_, obj_ = received_obj
        self.assertEqual(obj_, SimpleClass(1, 2))
    
    def test_transmit_data_class(self):
        dtcl: DataClass = DataClass(5, SimpleClass(1, 2))
        dtcl.add_one()
        obj = ({'a': 1, 'b': 2}, dtcl)
        with MeasureTime(f'{cen()} - codec'):
            with MeasureTime(f'{cen()} - serialization'):
                obj_id_result: ResultExistence[int] = self.rom_src.serialize(obj)
            
            self.assertTrue(obj_id_result)
            self.assertEqual(len(self.transport), 14)
            with MeasureTime(f'{cen()} - deserialization'):
                for item in self.transport:
                    item_type, item_info = item
                    self.assertIn(item_type, ('new_class', 'new_obj'))
                    if 'new_class' == item_type:
                        transmited_class_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_class(item_info)
                        self.assertTrue(transmited_class_result)
                    elif 'new_obj' == item_type:
                        transmited_obj_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_obj(item_info)
                        self.assertTrue(transmited_obj_result)
        
        received_object_id, received_obj = transmited_obj_result.value
        dict_, obj_ = received_obj
        self.assertEqual(obj_, dtcl)
    
    def test_transmit_data_class_pickle(self):
        dtcl: DataClass = DataClass(5, SimpleClass(1, 2))
        dtcl.add_one()
        obj = ({'a': 1, 'b': 2}, dtcl)
        with MeasureTime(f'{cen()} - codec'):
            with MeasureTime(f'{cen()} - serialization'):
                obj_id_result: ResultExistence[int] = self.rom_src_pickle.serialize(obj)
            
            self.assertTrue(obj_id_result)
            self.assertEqual(len(self.transport), 7)
            with MeasureTime(f'{cen()} - deserialization'):
                for item in self.transport:
                    item_type, item_info = pickle_loads(item)
                    self.assertIn(item_type, ('new_class', 'new_obj'))
                    if 'new_class' == item_type:
                        transmited_class_result: ResultExistence[Tuple[int, Any]] = self.rom_dest_pickle.deserialize_class(item_info)
                        self.assertTrue(transmited_class_result)
                    elif 'new_obj' == item_type:
                        transmited_obj_result: ResultExistence[Tuple[int, Any]] = self.rom_dest_pickle.deserialize_obj(item_info)
                        self.assertTrue(transmited_obj_result)
        
        received_object_id, received_obj = transmited_obj_result.value
        dict_, obj_ = received_obj
        self.assertEqual(obj_, dtcl)
    
    def test_transmit_data_class__raw_pickle(self):
        dtcl: DataClass = DataClass(5, SimpleClass(1, 2))
        dtcl.add_one()
        obj = ({'a': 1, 'b': 2}, dtcl)
        with MeasureTime(f'{cen()} - codec'):
            with MeasureTime(f'{cen()} - serialization'):
                pickled_obj = pickle_dumps(obj)
            
            with MeasureTime(f'{cen()} - deserialization'):
                unpickled_obj = pickle_loads(pickled_obj)
        
        self.assertEqual(unpickled_obj, obj)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AggregatorAppendFormatter', 'AggregatorView']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from math import ceil
import ttkbootstrap as ttkb
from ttkbootstrap.constants import INSERT as ttkb__INSERT, CURRENT as ttkb__CURRENT, SEL_FIRST as ttkb__SEL_FIRST, SEL_LAST as ttkb__SEL_LAST, SEL as ttkb__SEL, END as ttkb__END
from ttkbootstrap import Style
from typing import Callable, Hashable, Optional, Tuple, Any

from cengal.parallel_execution.coroutines.coro_standard_services.tkinter import *
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.fast_aggregator import *
from cengal.user_interface.gui.tkinter.components.read_only_text import *


class AggregatorAppendFormatter:
    def __init__(self, initial_text: str) -> None:
        self.initial_text: str = initial_text
        self._items_num: int = 0
        self.initiated: bool = False
    
    def reset(self):
        self._items_num = 0
        self.initiated = False
    
    def __call__(self, data: Any) -> Any:
        self._items_num += 1
        
        if self.initiated:
            return f'{data}'
        else:
            self.initiated = True
            return f'{current_interface().coro_id}. {self.initial_text}:\n{data}'


class AggregatorView(ttkb.Frame):
    def __init__(self, append_data: bool, default_auto_scroll: bool, aggregator_key: Hashable, updating_interval: float, data_formatter_func: Optional[Callable], width, height, *args, **kwargs):
        self.append_data: bool = append_data
        self.updating_interval: float = updating_interval
        self.aggregator_key: Hashable = aggregator_key
        self._default_data_formatter_func: Callable = lambda x: f'{x}'
        self.data_formatter_func: Optional[Callable] = data_formatter_func or self._default_data_formatter_func
        self._stop: bool = False
        self.updates_num: int = 0
        self.max_len: int = None
        super().__init__(*args, **kwargs)

        text_area_config = {
            'highlightcolor': Style.instance.colors.primary,
            'highlightbackground': Style.instance.colors.border,
            'highlightthickness': 1,
            'wrap': 'none',
        }
        if width is not None:
            text_area_config.update({
                'width': width,
            })
        
        if height is not None:
            text_area_config.update({
                'height': height,
            })
        
        self.text_area = ReadOnlyText(self, **text_area_config)
        self.text_area.pack(fill='both', expand='yes')

        self.text_area._text['yscrollcommand'] = self.yscroll
        self.text_area._vbar.config(command=self.yview)
        self.auto_scroll: bool = default_auto_scroll
        self.last_yscroll: Tuple[str, str] = None
        self.last_lines_num: int = None
        self.last_yview: Tuple[str] = None
        self.last_line: int = 0
        self.desired_line: int = 0
        self.selection_started = False
        self.selection_first = None
        self.selection_last = None
        self.selection = None
        self.previously_was_mouse_b1_double = False
        
        self.text_area._text.bind("<Button-3>", self.mouse_b3_down)
        if not append_data:
            self.text_area._text.bind("<<Selection>>", self.selection_changed)
            self.text_area._text.bind("<Double-Button-1>", self.mouse_b1_double)
            self.text_area._text.bind("<Button-1>", self.mouse_b1_down)
            self.text_area._text.bind("<B1-Motion>", self.mouse_b1_motion)
            self.text_area._text.bind("<ButtonRelease-1>", self.mouse_b1_up)
            self.text_area._text.bind("<Control-a>", self.ctrl_a)

    def ctrl_a(self, event):
        self.clear_selection()
        self.selection_first = self.text_area._text.index('@%s,%s' % (1, 0))
        self.selection_last = 'end'
        self.selection = (self.selection_first, self.selection_last)
        self.apply_selection()
        
        self.text_area._text.tag_remove(ttkb__INSERT, '1.0', 'end')
        self.text_area._text.tag_add(ttkb__INSERT, self.selection_last)
        self.text_area._text.mark_unset(ttkb__INSERT)
        self.text_area._text.mark_set(ttkb__INSERT, self.selection_last)
        return 'break'

    def clear_selection(self):
        self.text_area._text.tag_remove(ttkb__SEL, '1.0', 'end')
    
    def normalize_selection(self):
        if self.selection is not None:
            if 'end' == self.selection[0]:
                self.selection = (self.selection[1], self.selection[0])
            elif 'end' == self.selection[1]:
                pass
            else:
                sel_left = [int(x) for x in self.selection[0].split('.')]
                sel_right = [int(x) for x in self.selection[1].split('.')]
                if sel_left > sel_right:
                    self.selection = (self.selection[1], self.selection[0])
    
    def apply_selection(self):
        if self.selection is not None:
            if self.selection[0] == self.selection[1]:
                return
            
            self.normalize_selection()
            self.text_area._text.tag_remove(ttkb__SEL, '1.0', 'end')
            self.text_area._text.tag_add(ttkb__SEL, *self.selection)

    def mouse_b1_down(self, event):
        self.previously_was_mouse_b1_double = False
        self.selection_first = self.text_area._text.index('@%s,%s' % (event.x, event.y))
        self.selection_last = None
        self.clear_selection()

    def mouse_b1_up(self, event):
        selection_last = self.text_area._text.index('@%s,%s' % (event.x, event.y))
        if self.selection_first is not None:
            if self.selection_last is None:
                if self.selection_first != selection_last:
                    self.selection = (self.selection_first, selection_last)
                    self.apply_selection()
                    self.selection_first = None
                else:
                    self.selection = None
            else:
                self.selection = (self.selection_first, self.selection_last)
                self.apply_selection()
        
        self.text_area._text.tag_remove(ttkb__INSERT, '1.0', 'end')
        self.text_area._text.tag_add(ttkb__INSERT, selection_last)
        self.text_area._text.mark_unset(ttkb__INSERT)
        self.text_area._text.mark_set(ttkb__INSERT, selection_last)
        return 'break'

    def mouse_b1_motion(self, event):
        selection_last = self.text_area._text.index('@%s,%s' % (event.x, event.y))
        if self.selection_first is not None:
            self.selection = (self.selection_first, selection_last)
            self.apply_selection()
        
        return 'break'
    
    def mouse_b1_double(self, event):
        if self.previously_was_mouse_b1_double:
            self.selection_first = self.text_area._text.index('@%s,%s linestart' % (event.x, event.y))
            self.selection_last = self.text_area._text.index('@%s,%s lineend' % (event.x, event.y))
            self.selection = (self.selection_first, self.selection_last)
            self.apply_selection()
            self.previously_was_mouse_b1_double = False
        else:
            self.selection_first = self.text_area._text.index('@%s,%s wordstart' % (event.x, event.y))
            self.selection_last = self.text_area._text.index('@%s,%s wordend' % (event.x, event.y))
            self.selection = (self.selection_first, self.selection_last)
            self.apply_selection()
            self.previously_was_mouse_b1_double = True
        
        return 'break'

    def mouse_b3_down(self, event):
        self.text_area._text.mark_unset(ttkb__INSERT)
        selection = None
        if self.selection is not None:
            self.normalize_selection()
            selection = self.selection
        else:
            current_selection = self.text_area._text.index(ttkb__SEL)
            current_selection_first = self.text_area._text.index(ttkb__SEL_FIRST)
            current_selection_last = self.text_area._text.index(ttkb__SEL_LAST)
            if current_selection_first and current_selection_last:
                selection = (current_selection_first, current_selection_last)

        if selection is not None:
            self.text_area._text.clipboard_clear()
            text_content = self.text_area._text.get(selection[0], selection[1])
            self.text_area._text.clipboard_append(text_content)
            self.selection_last = None
            self.selection = None
            self.previously_was_mouse_b1_double = False
            self.clear_selection()
    
    def selection_changed(self, event):
        return

    def put(self, text: str):
        lines_num_before = self.index_to_line_number(self.text_area._text.index('end'))
        view_fractions_before = self.text_area._text.yview()
        line_before = int(ceil(lines_num_before * float(view_fractions_before[0])))
        tag_insert_before = self.text_area._text.index(ttkb__INSERT)
        tag_current_before = self.text_area._text.index(ttkb__CURRENT)
        
        if not self.append_data:
            self.clear_context()
        
        self.text_area._text.insert('end', text)
        
        if self.auto_scroll:
            self.text_area._text.yview(ttkb__END)

        if not self.append_data:
            self.text_area._text.tag_remove(ttkb__INSERT, '1.0', 'end')
            self.text_area._text.tag_add(ttkb__INSERT, tag_insert_before)
            self.text_area._text.tag_remove(ttkb__CURRENT, '1.0', 'end')
            self.text_area._text.tag_add(ttkb__CURRENT, tag_current_before)
            self.text_area._text.mark_unset(ttkb__INSERT)
            self.text_area._text.mark_set(ttkb__INSERT, tag_insert_before)
            self.apply_selection()
            
            lines_num_after = self.index_to_line_number(self.text_area._text.index('end'))
            view_fractions_after = self.text_area._text.yview()
            line_after = int(ceil(lines_num_after * float(view_fractions_after[0])))
            
            line_before = line_before or 1
            line_after = line_after or 1
            if line_after != line_before:
                movement = line_before - line_after
                self.text_area._text.yview_scroll(movement, 'units')
    
    def index_to_line_number(self, index: str):
        return int(index.split('.')[0])

    def clear_context(self):
        self.text_area._text.delete('1.0', 'end')

    def yscroll(self, *args):
        first, last = args
        lines_num = self.index_to_line_number(self.text_area._text.index('end')) - 1
        last_line_index = lines_num - 1
        first_visible_line_index = int(ceil(lines_num * float(first)))
        last_visible_line_index = int(ceil(lines_num * float(last))) - 1

        if self.auto_scroll:
            if (0 < first_visible_line_index) and (last_visible_line_index < last_line_index):
                self.auto_scroll = False
        else:
            if (0 <= last_line_index) and (0 <= last_visible_line_index) and (0 < first_visible_line_index) and (last_visible_line_index >= first_visible_line_index) and (last_visible_line_index == last_line_index):
                self.auto_scroll = True
        
        self.text_area._vbar.set(*args)

    def yview(self, *args):
        self.text_area._text.yview(*args)
    
    def start(self, wr: TkObjWrapper):
        wr.put_coro(self._updater)
    
    def stop(self):
        self._stop = True
    
    def _updater(self, i: Interface):
        while not self._stop:
            try:
                self.updates_num += 1
                if (self.max_len is not None) and (self.updates_num >= self.max_len):
                    if isinstance(self.data_formatter_func, AggregatorAppendFormatter):
                        self.updates_num = 1
                        self.data_formatter_func.reset()
                        auto_scroll_buff = self.auto_scroll
                        self.clear_context()
                        self.auto_scroll = auto_scroll_buff
                    
                data = i(FastAggregator, self.aggregator_key)
                if self.append_data:
                    text = '\n'.join([self.data_formatter_func(item) for item in data]) + '\n'
                else:
                    text = self.data_formatter_func(data[-1])
                    
                self.put(text)
            except KeyError:
                pass

            i(Sleep, self.updating_interval)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import select
from .net_io_abstract import *

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class IOMethodEpollLT(IOMethodBase):
    def __init__(self, interface: NetIOBase):
        super().__init__(interface)
        self.epoll = select.epoll()

    def loop_iteration(self, timeout=None):
        timeout = timeout or -1

        events = self.epoll.poll(timeout)
        for fileno, event in events:
            connection = self.interface.connection_by_fileno[fileno]

            if event & select.EPOLLIN:
                # Read available. We can try to read even even if an error occurred
                if ConnectionType.passive == connection.connection_info.connection_type:
                    self.interface.on_accept_connection(connection)
                else:
                    self.interface.on_read(connection)

            if event & select.EPOLLHUP:
                # Some error. Connection should be closed
                self.should_be_closed.add(connection.conn)
            elif event & select.EPOLLOUT:
                # Write available. We will not write data if an error occurred
                if ConnectionState.waiting_for_connection == connection.connection_state:
                    if not connection.conn.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR):
                        # Connected successfully:
                        self.interface.on_connected(connection)
                    else:
                        # Some connection error - will be closed:
                        self.should_be_closed.add(connection.conn)
                else:
                    self.interface.on_write(connection)

            self._close_all()

    def _close_all(self):
        should_be_closed = self.should_be_closed
        self.should_be_closed = set()
        for conn in should_be_closed:
            fileno = conn.fileno()
            if fileno in self.interface.connection_by_fileno:
                connection = self.interface.connection_by_fileno[fileno]
                self.interface.on_close(connection)  # self.remove_connection() should be run from inside of this
                #   callback
            else:
                self.remove_connection(conn)

    def destroy(self):
        self._close_all()
        self.epoll.close()

    def add_connection(self, conn: socket.socket):
        self.epoll.register(conn.fileno(), select.EPOLLIN)

    def remove_connection(self, conn: socket.socket):
        self.epoll.unregister(conn.fileno())

    def set__need_write(self, conn: socket.socket, state=True):
        if state:
            self.epoll.modify(conn.fileno(), select.EPOLLIN | select.EPOLLOUT)
        else:
            self.epoll.modify(conn.fileno(), select.EPOLLIN)

    def set__should_be_closed(self, conn: socket.socket):
        self.should_be_closed.add(conn)
